use std::str::FromStr;

use thiserror::Error;

use super::{seed_range::SeedRange, UnsignedNumber};

#[derive(Debug, PartialEq, Eq)]
pub struct ItemMapping {
    source_range: SeedRange,
    target: UnsignedNumber,
    range: UnsignedNumber,
}

impl ItemMapping {
    pub fn map_point(&self, to_map: UnsignedNumber) -> Option<UnsignedNumber> {
        let source_range = self.source_range;
        let start = source_range.start();
        let end = source_range.end();
        if start <= to_map && end >= to_map {
            let difference = to_map - start;
            let mapped = self.target + difference;
            Some(mapped)
        } else {
            None
        }
    }
}

#[derive(Debug, Error, PartialEq, Eq)]
#[error("Item mapping must be exactly 3 numbers")]
pub struct InvalidStrItemMapping;

impl FromStr for ItemMapping {
    type Err = InvalidStrItemMapping;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut splitted = s.split_whitespace().take(3).map(|to_parse| {
            to_parse
                .parse::<UnsignedNumber>()
                .map_err(|_| InvalidStrItemMapping)
        });
        match (splitted.next(), splitted.next(), splitted.next()) {
            (Some(Ok(target)), Some(Ok(source)), Some(Ok(range))) => {
                let inclusive_range = range.saturating_sub(1);
                let end = source + inclusive_range;
                let source_range = SeedRange::new(source, end);
                let parsed_self = Self {
                    source_range,
                    target,
                    range,
                };
                Ok(parsed_self)
            }
            (_, _, _) => Err(InvalidStrItemMapping),
        }
    }
}

#[cfg(test)]
mod testing {
    use crate::solutions::day5::{
        item_mapping::{InvalidStrItemMapping, ItemMapping},
        seed_range::SeedRange,
        UnsignedNumber,
    };

    #[test]
    fn from_str_item_mapping() {
        fn assert_case(input: &str, expected: Result<ItemMapping, InvalidStrItemMapping>) {
            let actual = input.parse();
            assert_eq!(expected, actual, "Input: {}", input);
        }
        assert_case(
            "20 10 22",
            Ok(ItemMapping {
                source_range: SeedRange::new(10, 31),
                target: 20,
                range: 22,
            }),
        );
        assert_case(
            "50 98 2",
            Ok(ItemMapping {
                source_range: SeedRange::new(98, 99),
                target: 50,
                range: 2,
            }),
        );
        assert_case(
            "52 50 48",
            Ok(ItemMapping {
                source_range: SeedRange::new(50, 97),
                target: 52,
                range: 48,
            }),
        );
        assert_case("a 10 22", Err(InvalidStrItemMapping));
        assert_case("10 10", Err(InvalidStrItemMapping));
    }
    #[test]
    fn should_map_point() {
        fn assert_case(point: UnsignedNumber, input: &str, expected: Option<UnsignedNumber>) {
            let mapping: ItemMapping = input.parse().unwrap();
            let actual = mapping.map_point(point);

            assert_eq!(
                expected, actual,
                "Point: {:?}\nMaping: {:?}",
                point, mapping
            );
        }

        assert_case(2, "4 2 2", Some(4));
        assert_case(0, "4 2 2", None);
        assert_case(4, "4 2 2", None);
        assert_case(79, "52 50 48", Some(81));
    }
}
