use std::collections::HashMap;

use crate::{
    iterations::{AdventIterator, Chunck},
    solutions::day5::parsing,
};

use super::{
    mapping_layer::MappingLayer,
    range_mapping::{MappedSeedRange, RangeMapping},
    seed_range::SeedRange,
    UnsignedNumber,
};

pub fn solve_task_2(input: &str) -> String {
    let (seeds, layers) = parsing::parse(input);
    let seed_ranges = combine_seeds_to_ranges(seeds);
    let mut min = UnsignedNumber::MAX;
    for next in seed_ranges {
        let loacal_min = simuluate_one_seed_range(next, &layers);
        min = min.min(loacal_min);
    }
    min.to_string()
}

fn simuluate_one_seed_range(seed_range: SeedRange, layers: &[MappingLayer]) -> UnsignedNumber {
    let mut current_branches = vec![seed_range];

    for next in layers {
        let mut cache = CacheOfMappedRanges::new();
        let ranges = next.ranges();
        #[cfg(debug_assertions)]
        {
            let mut sorted: Vec<UnsignedNumber> = current_branches
                .iter()
                .copied()
                .map(|e| e.start())
                .collect();
            let unsorted = sorted.clone();
            sorted.sort();
            assert_eq!(unsorted, sorted);
        }

        for next_seed_range in current_branches {
            map_sequence(next_seed_range, ranges, &mut cache);
        }

        current_branches = cache
            .into_iter()
            .map(|(start, end)| SeedRange::new(start, end))
            .collect();
        current_branches.sort_by_key(|not_the_key| not_the_key.start());
    }

    let min_location = current_branches
        .into_iter()
        .min_by_key(|seed_range| seed_range.start())
        .unwrap();
    min_location.start()
}

type CacheOfMappedRanges = HashMap<UnsignedNumber, UnsignedNumber>;
fn map_sequence(
    seed_range: SeedRange,
    mapping: &[RangeMapping],
    found_mapping: &mut CacheOfMappedRanges,
) -> Vec<SeedRange> {
    fn add_to_cache(range: SeedRange, cache: &mut CacheOfMappedRanges) {
        let value = range.end();
        cache
            .entry(range.start())
            .and_modify(|current_max| *current_max = (*current_max).max(value))
            .or_insert(value);
    }
    #[cfg(debug_assertions)]
    {
        let last = mapping.last().unwrap();
        assert!(last.source() == last.target(), "last: {:?}", last);
        let to_test_for_sorting = mapping
            .into_iter()
            .map(|e| e.source())
            .collect::<Vec<UnsignedNumber>>();
        let is_sorted = crate::sequences::is_sorted(to_test_for_sorting.as_slice());
        assert!(is_sorted);
    }
    let mut new_branches = Vec::new();
    let mut currend_seed_range = seed_range;
    for next in mapping {
        match next.map_range(currend_seed_range) {
            MappedSeedRange::OutsideFromLeft => (),
            MappedSeedRange::Completely(done) => {
                add_to_cache(done, found_mapping);
                new_branches.push(done);
                return new_branches;
            }
            MappedSeedRange::Partially { mapped, left } => {
                add_to_cache(mapped, found_mapping);
                new_branches.push(mapped);
                currend_seed_range = left;
            }
        }
    }
    unreachable!();
}

fn combine_seeds_to_ranges(seeds: Vec<UnsignedNumber>) -> Vec<SeedRange> {
    seeds
        .into_iter()
        .into_chunks::<2>()
        .map(|next_chunk| match next_chunk {
            Chunck::Next([start, range]) => {
                let end = start + range - 1;
                SeedRange::new(start, end)
            }
            Chunck::Rest(_) => panic!("Uneven number of seeds"),
        })
        .collect()
}

#[cfg(test)]
mod testing {
    use crate::solutions::day5::solve_task_2;

    #[test]
    fn task2() {
        let actual = solve_task_2(include_str!("day5_example_input.txt"));
        assert_eq!("46", actual);
    }
}
